#include <kr2_program_api/api_v1/cbun/xmlrpc/xmlrpc_server.h>

#include "domain/program.h"
#include "domain/group.h"
#include <iostream>
#include <functional>
#include <set>
#include <kr2_program_api/api_v1/utils/xmlrpc/base.h>
#include "spdlog/spdlog.h"

namespace spin
{
    using UUID = boost::uuids::uuid;

    class SafetyShieldReleaseBit : public kr2_xmlrpc::Method
    {

    public:
        SafetyShieldReleaseBit(std::function<void()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            callback_();
            return kr2_xmlrpc::Value::Int(0);
        }

    private:
        std::function<void()> callback_;
    };

    class SafetyShieldLockBit : public kr2_xmlrpc::Method
    {
    public:
        SafetyShieldLockBit(std::function<void()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            callback_();
            return kr2_xmlrpc::Value::Int(0);
        }

    private:
        std::function<void()> callback_;
    };

    class IsBridgeConnected : public kr2_xmlrpc::Method
    {
    public:
        IsBridgeConnected(std::function<bool()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            // This method should actually return a boolean value, but the Value class' boolean type is bugged,
            // or maybe the client is bugged. Anyway, the client allways receives a boolean false, so we return an int instead.
            // The bug has been reported to Kassow.
            return kr2_xmlrpc::Value::Int(callback_());
        }

    private:
        std::function<bool()> callback_;
    };

    class IsToolConnected : public kr2_xmlrpc::Method
    {
    public:
        IsToolConnected(std::function<bool()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            return kr2_xmlrpc::Value::Int(callback_());
        }

    private:
        std::function<bool()> callback_;
    };

    class Programs : public kr2_xmlrpc::Method
    {
    public:
        Programs(std::function<std::set<Program>()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            try
            {
                std::set<Program> programs = callback_();
                std::vector<kr2_xmlrpc::Value> programsVector;
                for (auto program : programs)
                {
                    programsVector.push_back(toXmlrpc(program));
                }
                return kr2_xmlrpc::Value::Array(programsVector);
            }
            catch (const std::exception &e)
            {
                SPDLOG_ERROR("Programs::execute: {}", e.what());
                return kr2_xmlrpc::Value::Null();
            }
        }

    private:
        std::function<std::set<Program>()> callback_;
        kr2_xmlrpc::Value toXmlrpc(Program program)
        {
            // return kr2_xmlrpc::Value::String(boost::uuids::to_string(program.uuid_));

            // TODO: This does not work, Kassow will have to fix it on their side.

            std::map<std::string, kr2_xmlrpc::Value> programMap;
            programMap.emplace("uuid", kr2_xmlrpc::Value::String(boost::uuids::to_string(program.uuid_)));
            programMap.emplace("name", kr2_xmlrpc::Value::String(program.name_));
            programMap.emplace("reference_depth_mm", kr2_xmlrpc::Value::Double(program.reference_depth_.toMillimeters()));
            programMap.emplace("start_depth_mm", kr2_xmlrpc::Value::Double(program.start_depth_.toMillimeters()));
            programMap.emplace("initial_safety_shield_position_mm", kr2_xmlrpc::Value::Double(program.initial_safety_shield_position_.toMillimeters()));
            programMap.emplace("is_force_controlled", kr2_xmlrpc::Value::Int(program.is_force_controlled_ ? 1 : 0)); // TODO: This is a hack, Kassow should fix this.

            return kr2_xmlrpc::Value::Struct(programMap);
        }
    };

    class Groups : public kr2_xmlrpc::Method
    {
    public:
        Groups(std::function<std::set<Group>()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            std::set<Group> groups = callback_();
            std::vector<kr2_xmlrpc::Value> groups_vector;
            for (auto group : groups)
            {
                std::cout << "Group: " << group << std::endl;
                groups_vector.push_back(toXmlrpc(group));
            }
            return kr2_xmlrpc::Value::Array(groups_vector);
        }

    private:
        std::function<std::set<Group>()> callback_;
        kr2_xmlrpc::Value toXmlrpc(Group group)
        {
            return kr2_xmlrpc::Value::String(group.name_);

            std::map<std::string, kr2_xmlrpc::Value> group_map;
            group_map.at("uuid") = kr2_xmlrpc::Value::String(boost::uuids::to_string(group.uuid_));
            group_map.at("name") = kr2_xmlrpc::Value::String(group.name_);
            std::vector<kr2_xmlrpc::Value> programs_vector;
            for (auto uuid : group.programUuids)
            {
                programs_vector.push_back(kr2_xmlrpc::Value::String(boost::uuids::to_string(uuid)));
            }
            group_map.at("programUUIDs") = kr2_xmlrpc::Value::Array(programs_vector);

            return kr2_xmlrpc::Value::Struct(group_map);
        }
    };

    class SafetyShieldDoHoming : public kr2_xmlrpc::Method
    {
    public:
        SafetyShieldDoHoming(std::function<void()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            callback_();
            return kr2_xmlrpc::Value::Int(0);
        }

    private:
        std::function<void()> callback_;
    };

    class AcknowledgeErrors : public kr2_xmlrpc::Method
    {
    public:
        AcknowledgeErrors(std::function<void()> callback) : callback_(callback) {}

        kr2_xmlrpc::Value execute(const kr2_xmlrpc::Params &params) override
        {
            callback_();
            return kr2_xmlrpc::Value::Int(0);
        }

    private:
        std::function<void()> callback_;
    };

} // namespace spin
